/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.bindings;

import static org.objectweb.fractal.cecilia.adl.bindings.BindingDecorationUtil.getClientInterface;
import static org.objectweb.fractal.cecilia.adl.bindings.BindingDecorationUtil.getServerInterface;
import static org.objectweb.fractal.cecilia.adl.bindings.BindingDecorationUtil.isUsedAtStart;
import static org.objectweb.fractal.cecilia.adl.components.ComponentDecorationUtil.setStartOrder;
import static org.objectweb.fractal.cecilia.adl.interfaces.InterfaceDecorationUtil.getContainer;

import java.util.AbstractSet;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.bindings.Binding;
import org.objectweb.fractal.adl.bindings.BindingContainer;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.util.FractalADLLogManager;
import org.objectweb.fractal.cecilia.adl.components.ComponentDecorationUtil;

/**
 * A delegating loader that checks, for each composite, the graph defined by
 * sub-components and bindings, in order to checks that a correct start order
 * exists.
 * <p>
 * To do so, it builds the graph defined by the <i>must be starter before</i>
 * binary relation, and checks their is no cycle in this graph. More precisly,
 * the <i>must be starter before</i> binary relation is defined as follows :
 * component <code>C1</code> must be started before component <code>C2</code>
 * if and only if component C1 is transitivly accessible by a "used-at-start"
 * client interface of component <code>C2</code>. By "transitivly
 * accessible", we mean that it exist a walk in the binding graph from
 * <code>C2</code> to <code>C1</code>.
 * <p>
 * Indeed for a given component, every components transitivly accessible trough
 * a "used-at-start" client interface must be started before it since they may
 * be invoked during its start.
 * <p>
 * The alogorithm, implemented in this checker, allows reflexive <i>must be
 * starter before</i> relation (i.e. It allows relation of the form
 * <code>C1</code> must be started before <code>C1</code>). In this case a
 * message is printed to warn that a component may be invoked during its own
 * start.
 * <p>
 * Finnally this component adds a
 * {@link ComponentDecorationUtil#SUB_COMPONENT_START_ORDER} decoration to each
 * composite.
 */
public class BindingCycleDetectorLoader extends AbstractLoader {

  // --------------------------------------------------------------------------
  // Implementation of the Loader interface
  // --------------------------------------------------------------------------

  public Definition load(final String name, final Map<Object, Object> context)
      throws ADLException {
    final Definition d = clientLoader.load(name, context);
    checkNode(d, context);
    return d;
  }

  // --------------------------------------------------------------------------
  // Checking methods
  // --------------------------------------------------------------------------

  private void checkNode(final Object node, final Map<Object, Object> context)
      throws ADLException {
    if (node instanceof ComponentContainer) {
      for (final Component comp : ((ComponentContainer) node).getComponents()) {
        checkNode(comp, context);
      }
    }
    if (node instanceof BindingContainer
        && ((BindingContainer) node).getBindings().length > 0) {

      // create a map associating sub component node to the list of the bindings
      // it is the client part. This map defines a binary relation on the set of
      // components (this binary relation is defined by bindings).
      final Map<Object, List<Binding>> bindingGraph = new HashMap<Object, List<Binding>>();

      for (final Binding binding : ((BindingContainer) node).getBindings()) {

        final Object client = getContainer(getClientInterface(binding));
        List<Binding> bindings = bindingGraph.get(client);
        if (bindings == null) {
          bindings = new ArrayList<Binding>();
          bindingGraph.put(client, bindings);
        }
        bindings.add(binding);
      }

      // build the transitive closure of the bindingGraph. i.e. create a graph
      // associating a given component to the set of every component
      // transitively accessible through the bindingGraph.
      final Map<Object, Set<Object>> transitiveBindingGraph = new HashMap<Object, Set<Object>>();
      for (final Object component : bindingGraph.keySet()) {
        final Set<Object> transitivelyAccessibleComponents = new IdentityHashSet<Object>();
        buildTransitiveGraph(bindingGraph, transitivelyAccessibleComponents,
            component);
        // put the key-set of transitivelyAccessibleComponents in
        // transitiveBindingGraph
        transitiveBindingGraph.put(component, transitivelyAccessibleComponents);
      }

      // create a graph associating each component to the set of every
      // components transitively accessible through its "used-at-start"
      // bindings. This graph contains the start dependancies (i.e. the set of
      // component that must be started before a given one). This graph must not
      // contain cycle in order to be able to find a correct start order.
      final Map<Object, Set<Object>> startDependencyGraph = new HashMap<Object, Set<Object>>();

      // for each entry in the binding graph
      for (final Map.Entry<Object, List<Binding>> graphEntry : bindingGraph
          .entrySet()) {
        Set<Object> dependancy = null;

        final Object component = graphEntry.getKey();

        // for each bindings, for which component is the client
        for (final Binding binding : graphEntry.getValue()) {

          // consider only "used-at-start" bindings
          if (!isUsedAtStart(binding)) continue;

          // get the server side component of this binding
          final Object serverComponent = getContainer(getServerInterface(binding));

          // instantiate dependancy set lazily
          if (dependancy == null) dependancy = new IdentityHashSet<Object>();

          // add the serverComponnet as a start dependancy of the current
          // component.
          dependancy.add(serverComponent);

          // add every components that are transitivly accessible through
          // serverComponent (if any).
          final Set<Object> transitivelyAccessibleComponents = transitiveBindingGraph
              .get(serverComponent);
          if (transitivelyAccessibleComponents != null)
            dependancy.addAll(transitiveBindingGraph.get(serverComponent));
        }

        if (dependancy != null) {
          startDependencyGraph.put(component, dependancy);
        }
      }

      // find cycle in the startDependencyGraph.
      detectCycle(node, startDependencyGraph);

      // set startOrder
      createStartOrder((ComponentContainer) node, startDependencyGraph);
    }
  }

  private void buildTransitiveGraph(
      final Map<Object, List<Binding>> bindingGraph,
      final Set<Object> transitivelyAccessibleComponents, final Object component) {
    final List<Binding> bindings = bindingGraph.get(component);
    if (bindings == null) return;
    for (final Binding binding : bindings) {
      final Object serverComponent = getContainer(getServerInterface(binding));
      if (transitivelyAccessibleComponents.add(serverComponent)) {
        // serverComponent added in the set, call recursivly
        buildTransitiveGraph(bindingGraph, transitivelyAccessibleComponents,
            serverComponent);
      }
    }
  }

  private void detectCycle(final Object container,
      final Map<Object, Set<Object>> graph) throws StartDependencyCycle {
    final Set<Object> checkedNodes = new IdentityHashSet<Object>();
    for (final Object node : graph.keySet()) {
      final Set<Object> path = new IdentityHashSet<Object>();
      path.add(node);
      doDetectCycle(container, graph, checkedNodes, path, node);
    }
  }

  private GraphCycle doDetectCycle(final Object container,
      final Map<Object, Set<Object>> graph, final Set<Object> checkedNodes,
      final Set<Object> path, final Object node) throws StartDependencyCycle {
    // if node has already been checked, returns
    if (checkedNodes.contains(node)) return null;

    final Set<Object> nodesInRelation = graph.get(node);
    if (nodesInRelation != null) {
      // for each node in realation with the given node
      for (final Object n : nodesInRelation) {
        // ignore "one node cycle". i.e. a node that is in relation with itself.
        if (n == node) {
          // TODO allows to specify in the ADL that a server interface can be
          // safely invoked (without warning) during the start.
          if (FractalADLLogManager.STEP.isLoggable(Level.WARNING)) {
            FractalADLLogManager.STEP
                .warning("WARNING : Server interfaces of component \""
                    + ((Component) node).getName()
                    + "\" may be invoked during its own start (component defined at: "
                    + ((Node) node).astGetSource() + ")");
          }
          continue;
        }

        if (!path.add(n)) {
          // cycle detected !
          return new GraphCycle(n);
        } else {
          final GraphCycle result = doDetectCycle(container, graph,
              checkedNodes, path, n);
          path.remove(n);
          // if a cycle has been detected in recursive call
          if (result != null) {
            result.addNode(n);
            if (node == result.inCycleNode) {
              // the full cycle list is built. throw an exception
              throw new StartDependencyCycle(result.cycle, container);
            }
            return result;
          }
        }
      }
    }

    // node has been correctly checked. add to checkedNode
    checkedNodes.add(node);
    return null;
  }

  private void createStartOrder(final ComponentContainer container,
      final Map<Object, Set<Object>> graph) {
    final List<Component> startList = new ArrayList<Component>(container
        .getComponents().length);
    final IdentityHashSet<Component> inList = new IdentityHashSet<Component>();
    for (final Object node : graph.keySet()) {
      doBuildStartOrder(container, graph, (Component) node, startList, inList);
    }
    final Component[] subComponents = container.getComponents();
    if (subComponents.length != startList.size()) {
      // some components are not in the start list. They can be started at any
      // time, add them at the end of the list
      for (final Component c : subComponents) {
        if (!inList.contains(c)) {
          startList.add(c);
        }
      }
    }
    setStartOrder(container, startList);
  }

  private void doBuildStartOrder(final ComponentContainer container,
      final Map<Object, Set<Object>> graph, final Component node,
      final List<Component> startList, final Set<Component> inList) {
    if (inList.contains(node)) return;
    final Set<Object> nodesInRelation = graph.get(node);
    if (nodesInRelation != null) {
      // for each node in realation with the given node
      for (final Object n : nodesInRelation) {
        if (n == node || n == container) continue;
        doBuildStartOrder(container, graph, (Component) n, startList, inList);
      }
    }
    if (inList.add(node)) {
      startList.add(node);
    }
  }

  private static final class GraphCycle {
    /** List of node involved in the cycle. */
    private final List<Object> cycle;
    /** The node on which the cycle has been detected. */
    private final Object       inCycleNode;

    private GraphCycle(final Object node) {
      inCycleNode = node;
      cycle = new ArrayList<Object>();
      cycle.add(node);
    }

    private void addNode(final Object b) {
      cycle.add(b);
    }
  }

  private static class IdentityHashSet<E> extends AbstractSet<E> {

    private final IdentityHashMap<E, E> storage = new IdentityHashMap<E, E>();

    @Override
    public boolean add(final E o) {
      return storage.put(o, o) == null;
    }

    @Override
    public boolean remove(final Object o) {
      return storage.remove(o) != null;
    }

    @Override
    public Iterator<E> iterator() {
      return storage.keySet().iterator();
    }

    @Override
    public int size() {
      return storage.size();
    }

  }
}
